"""Provides a visual panel (L{qubx.faces.Face}) to optimize single-molecule rate constants.

Uses math from U{qubx_single_molecule<http://www.qub.buffalo.edu/src/qub-express/api/qubx_single_molecule>} 

/* Copyright 2008-2014 Research Foundation State University of New York   */

/* This file is part of QUB Express.                                      */

/* QUB Express is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Express is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Express program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

"""

import itertools
import gc
import os
import sys
import time
import traceback

from qubx.data_types import *
from qubx.util_types import *
from qubx.faces import *
from qubx.GTK import *
import qubx.notebook
import qubx.optimize
import qubx.optimizeGTK
import qubx.pyenv
from qubx.settings import Propertied, Property
import qubx.tree
from qubx.trialset import *
from qubx.model import *
import qubx.fast.model
from qubx.fast.fast_utils import buffer_as_array_of_int, buffer_as_array_of_double, buffer_as_matrix_of_int, buffer_as_matrix_of_double
from qubx.task import *
from max_subi_ll import *
from max_inter_ll import inter_ll_or_0

import scipy.optimize
from numpy import *


MAXITER = 300
LBL_TDEAD = 'Tdead (ms)'

OPT_SIMPLEX = 0
OPT_DFP     = 1

def cl_dimension(Ns):
    dim = 4
    while dim < Ns:
        dim *= 2
    return dim

class ModelPrep(Anon):
    """Represents a model ready for computation."""
    def __init__(self, model=None, modeltree=None): # one or the other, not both; modeltree is more efficient
        Anon.__init__(self, source=model, modeltree=modeltree or model.as_tree(), model=QubModel())
        self.model.from_tree(self.modeltree)
        self.K0, self.K1, self.K2, L, V, P = ModelToRateMatricesP(self.model)
    def set_rates2(self, K0, K1, K2):
        self.K0, self.K1, self.K2 = K0, K1, K2
        K012ToModel(K0, K1, K2, self.model)

class DataPrep(Anon):
    """Represents data ready for computation."""
    def __init__(self, segs, ws):
        Anon.__init__(self, segs=segs, ws=ws)
        
class Workspace(Anon):
    """Holds data, model, and results for one file."""
    pass

def get_data_prep(segs, model_prep, wait=True, receiver=lambda dp: None):
    if wait:
        return qubx.pyenv.call_async_wait(lambda rcv: get_data_prep(segs, model_prep, wait=False, receiver=rcv))
    else:
        QubX = qubx.global_namespace.QubX
        view_of_file = dict([(view.file, view) for view in QubX.Data.views])
        view_segm = []
        vs_by_file = {}
        for seg in segs:
            if not (seg.file in vs_by_file):
                vs = (view_of_file[seg.file], [])
                view_segm.append(vs)
                vs_by_file[seg.file] = vs
            vs_by_file[seg.file][1].append(seg)
        if not view_segm:
            gobject.idle_add(receiver, DataPrep(segs, []))
            return
        workspaces = []
        data_prep = DataPrep(segs, workspaces)
        # stimulus idl could take a while, so its results come back on gobject.idle_add.
        # we'll use idle_add to alternate recursively between build_workspace and receive_stim,
        # until all workspaces are built.
        def build_workspace():
            i = len(workspaces)
            view, segm = view_segm[i]
            ws = Workspace(view=view)
            workspaces.append(ws)
            if view:
                for i in xrange(view.file.constants.size):
                    if LBL_TDEAD == view.file.constants.get(i, 'Name'):
                        ws.td_ms = view.file.constants.get(i, 'Value')
                        break
                else:
                    ws.td_ms = view.file.sampling * 1e3
                    view.file.constants.append({'Name':LBL_TDEAD, 'Value':ws.td_ms})
                    view.file.constants.select(view.file.constants.size-1, 'Value')
            
                ws.sampling_ms = view.file.sampling * 1e3
                ws.sig_names = [view.file.signals.get(segm[0].signal, 'Name')]
                ws.signals = [ [] ]
                for seg in segm:
                    ff, ll, cc, dd = QubX.DataSource.get_idealization(seg, mark_excluded=True, get_fragments=True, get_durations=True)
                    aa = view.file.ideal[segm[0].signal].seg[seg.index].amp
                    ws.signals[0].append( (cc, dd, aa) )
                QubX.Modeling.Stimulus.call_with_stimulus_idl(receive_stim, [(seg.f, seg.l) for seg in segm], datafile=view.file, model=model_prep.source or model_prep.model)
            else:
                receive_stim([], [])
        def receive_stim(stim_names, stimidl):
            i = len(workspaces) - 1
            ws = workspaces[i]
            ws.sig_names.extend(stim_names)
            for segs_flcda in stimidl:
                ws.signals.append( [(cc, dd, aa) for ff,ll,cc,dd,aa in segs_flcda] )
            try:
                ws.segs, ws.mcls, ws.scls = ProcessSignals(ws.td_ms, ws.signals)
            except:
                traceback.print_exc()
                ws.segs, ws.mcls, ws.scls = [], [], []
            del ws.signals
            if len(workspaces) == len(view_segm):
                gobject.idle_add(receiver, data_prep) # exit from recursion, return workspaces via idle_add
            else:
                gobject.idle_add(build_workspace)
        build_workspace()

class MultiplexedSignals(Requestable):
    def __init__(self, datasrc, Data, stimulus):
        Requestable.__init__(self)
        self.datasrc = datasrc
        self.Data = Data
        self.stimulus = stimulus
        self.__ref = Reffer()
        self.__use_group = None
        self.__data = None
        self.OnChangeTdead = WeakEvent()
        self.datasrc.OnChangeIdealization += self.__ref(self.__onChangeIdl)
        self.Data.table.OnInsert += self.__ref(self.__onChangeFileList)
        self.Data.table.OnRemoved += self.__ref(self.__onChangeFileList)
        self.Data.table.OnSet += self.__ref(self.__onSetFileList)
        self.Data.OnSwitch += self.__ref(self.__onSwitchDataFile)
        self.__datafile = None
        self.__onSwitchDataFile(self.Data, self.Data.file)
        self.stimulus.OnInsert += self.__ref(self.__onChangeStimulus)
        self.stimulus.OnChange += self.__ref(self.__onChangeStimulus)
        self.stimulus.OnRemoving += self.__ref(self.__onChangeStimulus)
        self.invalidate()
    tdead = property(lambda self: self.get_tdead(), lambda self, x: self.set_tdead(x))
    def get_tdead(self):
        if self.__datafile:
            for i in xrange(self.__datafile.constants.size):
                if LBL_TDEAD == self.__datafile.constants.get(i, 'Name'):
                    return self.__datafile.constants.get(i, 'Value')
            td_ms = self.__datafile.sampling * 1e3
            self.__datafile.constants.append({'Name':LBL_TDEAD, 'Value':td_ms})
            return td_ms
    def set_tdead(self, x, scriptable=False):
        if self.__datafile:
            qubx.pyenv.env.OnScriptable('QubX.Data.file.constants[QubX.Data.file.constants.index(%s), "Value"] = %s' % (repr(LBL_TDEAD), repr(x)))
            cc = self.__datafile.constants
            try:
                cc[cc.index(LBL_TDEAD), 'Value'] = x
            except KeyError:
                print 'surprise!'
                cc.append({'Name':LBL_TDEAD, 'Value':x})
    def set_use_group(self, x):
        if self.__use_group != x:
            self.__use_group = x
            self.invalidate()
    use_group = property(lambda self: self.__use_group, lambda self, x: self.set_use_group(x))
    def __onChangeIdl(self, *args):
        self.invalidate() # TODO: on scroll, if use_group != None, no invalidate
    def __onChangeFileList(self, *args):
        if self.__use_group != None:
            self.invalidate()
    def __onSetFileList(self, i, field_name, val, prev, undoing):
        if (self.__use_group != None) and (field_name == 'Group'):
            self.invalidate()
    def __onChangeStimulus(self, *args):
        self.invalidate()
    def __onSwitchDataFile(self, Data, file):
        if file == self.__datafile: return
        if self.__datafile:
            self.__datafile.constants.OnSet -= self.__ref(self.__onSetConstant)
        self.__datafile = file
        if self.__datafile:
            self.__datafile.constants.OnSet += self.__ref(self.__onSetConstant)
        self.OnChangeTdead()
    def __onSetConstant(self, r, field, val, prev, undoing):
        if self.__datafile.constants[r,'Name'] == LBL_TDEAD:
            self.invalidate()
            self.OnChangeTdead()
    def get(self, receiver):
        if self.__use_group != None:
            view_segm = self.datasrc.get_segmentation_files(group=self.__use_group)
        else:
            view_segm = [(self.Data.view, self.datasrc.get_segmentation())]
        segs = []
        for view, segm in view_segm:
            segs.extend(segm)
        get_data_prep(segs, self.model_prep, wait=False, receiver=receiver)


QUBX_SINGLE_RATES_METHODS = ["MIL", "MSL"]

@Propertied(Property('use_Peq', True, "True to start each segment at equilibrium; False to use States:Pr"),
            Property('use_group', False, "True to fit a group of data files", in_tree=False),
            Property('use_group_index', 1, 'Desired Group in Data table'),
            Property('useGPU', True, "True to use OpenCL when available"),
            Property("method", QUBX_SINGLE_RATES_METHODS[0], "choice of algorithm in %s" % ', '.join(QUBX_SINGLE_RATES_METHODS)),
            Property("tdead_in_samples", True, "False to display Tdead in milliseconds")
            )
class RatesFace(Face):
    def __init__(self):
        Face.__init__(self, 'MILRates', 'QubX.Modeling.MILRates')
        self.__ref = Reffer()
        self.propertied_connect_settings('Rates')

        self.set_size_request(200, 50)
        self.QubX = qubx.pyenv.env.globals['QubX']
        self.QubX.OnQuit += self.__ref(self.qubx_finalize)
        self.models = self.QubX.Models
        self.models.OnSwitch += self.__ref(self.__onSwitchModel), 'Rates.onSwitchModel'
        self.robot = Robot('MILRates')
        self.robot.OnException += self.__ref(self.__onException)
        self.robot.do(self.__robot_setup_cl_context)
        self.plexig = MultiplexedSignals(self.QubX.DataSource, self.QubX.Data, self.QubX.Modeling.Stimulus)
        self.plexig.use_group = self.use_group_index if self.use_group else None
        self.plexig.OnInvalidate += self.__ref(self.__onInvalidateData)
        self.plexig.OnChangeTdead += self.__ref(self.__onChangeTdead)
        self.model_prep = Product(self.__get_model_prep)
        self.model_prep.OnInvalidate += self.__ref(self.__onInvalidateModel)
        self.rates = Product(self.__get_rates)
        self.rates.OnInvalidate += self.__ref(self.__onInvalidateModel)
        self.LL = 0.0
        self.iterations = 0
        self.grads = []
        self.__on_trial = None

        qubx.optimize.setup_opt("qubx_single_rates_optimize",
                                {"max step" : 0.1})
        
        self.OnLL = WeakEvent()
        self.OnIteration = WeakEvent()
        self.OnOptimized = WeakEvent()

        qubx.notebook.Notebook.register_auto('Rates_eval.Messages', 'Messages, on MILRates: Get LL', False)
        qubx.notebook.Notebook.register_auto('Rates_eval.Model.Picture', 'Model picture, on MILRates: Get LL', False)
        qubx.notebook.Notebook.register_auto('Rates_eval.Model.PictureNO', 'Model picture, no overlays, on MILRates: Get LL', False)
        qubx.notebook.Notebook.register_auto('Rates_eval.Data.Picture', 'Data picture, on MILRates: Get LL', False)
        qubx.notebook.Notebook.register_auto('Rates_eval.Data.PictureNO', 'Data picture, no overlays, on MILRates: Get LL', False)
        qubx.notebook.Notebook.register_auto('Rates_eval.DurHist.Charts', 'DurHist charts, on MILRates: Get LL', False)
        qubx.notebook.Notebook.register_auto('Rates_eval.DurHist.TimeConst', 'DurHist time constants, on MILRates: Get LL', False)
        qubx.notebook.Notebook.register_auto('Rates_eval.Model.Rates', 'Rates table, on MILRates: Optimize', False)

        qubx.notebook.Notebook.register_auto('Rates_opt.Messages', 'Messages, on MILRates: Optimize', True)
        qubx.notebook.Notebook.register_auto('Rates_opt.Model.Picture', 'Model picture, on MILRates: Optimize', True)
        qubx.notebook.Notebook.register_auto('Rates_opt.Model.PictureNO', 'Model picture, no overlays, on MILRates: Optimize', False)
        qubx.notebook.Notebook.register_auto('Rates_opt.Data.Picture', 'Data picture, on MILRates: Optimize', True)
        qubx.notebook.Notebook.register_auto('Rates_opt.Data.PictureNO', 'Data picture, no overlays, on MILRates: Optimize', False)
        qubx.notebook.Notebook.register_auto('Rates_opt.DurHist.Charts', 'DurHist charts, on MILRates: Optimize', True)
        qubx.notebook.Notebook.register_auto('Rates_opt.DurHist.TimeConst', 'DurHist time constants, on MILRates: Optimize', True)
        qubx.notebook.Notebook.register_auto('Rates_opt.Model.Rates', 'Rates table, on MILRates: Optimize', False)
        qubx.notebook.Notebook.register_auto('Rates_opt.Model.Constraints', 'Kinetic Constraints table, on MILRates: Optimize', True)

        cols = pack_item(gtk.HBox(), self, expand=True)
        ctrls = pack_item(gtk.VBox(), cols, expand=True)
        row = pack_item(gtk.HBox(), ctrls)
        self.chkPeq = pack_check('start at equilibrium prob.   ', row)
        self.propertied_connect_check('use_Peq', self.chkPeq)
        pack_space(row, expand=True)
        self.chkUseSingle = pack_radio('Single file', row)
        self.chkUseGroup = pack_radio('Multi-file: Group', row, group=self.chkUseSingle)
        self.chkUseGroup.set_tooltip_text('Assign Groups in the Data table')
        self.propertied_connect_radios('use_group', [(False, self.chkUseSingle),
                                                     (True, self.chkUseGroup)])
        self.txtUseGroup = pack_item(qubx.GTK.NumEntry(self.use_group_index, acceptIntGreaterThan(-1), width_chars=3), row)
        self.propertied_connect_NumEntry('use_group_index', self.txtUseGroup)
        pack_space(row, expand=True)

        pack_label('', row).set_markup('t<sub>dead</sub>:')
        td = (1e-3*self.plexig.tdead/self.QubX.Data.file.sampling) if self.tdead_in_samples else self.plexig.tdead
        self.txtTdead = pack_item(qubx.GTK.NumEntry(td, acceptFloatGreaterThanOrEqualTo(0.0), '%.2g', width_chars=5), row)
        self.txtTdead.set_tooltip_text("""Note for multi-file: this is Tdead for the onscreen Data file only.
To edit another file's Tdead, first select it in the Data panel.""")
        self.txtTdead.OnChange += self.__ref(self.__onEditTdead)
        self.chkTdeadMS = pack_radio('[ms]', row)
        self.chkTdeadSamples = pack_radio('[samples]', row, group=self.chkTdeadMS)
        self.propertied_connect_radios('tdead_in_samples', [(False, self.chkTdeadMS), (True, self.chkTdeadSamples)])
        pack_space(row, expand=True)

        self.txtStatus = gtk.TextView()
        self.outStatus = TextViewAppender(self.txtStatus)
        self.txtStatus.set_editable(False)
        SetFixedWidth(self.txtStatus)
        pack_scrolled(self.txtStatus, ctrls, expand=True)
        self.outStatus.write("""Based on "Estimating single-channel kinetic parameters from idealized patch-clamp data containing missed events," Qin F, Auerbach A, Sachs F, Biophys J. 1996 Jan; 70(1):264-80
and "Maximum likelihood estimation of aggregated Markov processes," Qin F, Auerbach A, Sachs F, Proc Biol Sci. 1997 Mar 22; 264(1380):375-83.

""")

        ctrls = pack_item(gtk.VBox(), cols)
        row = pack_item(gtk.HBox(), ctrls)
        pack_label('LL:', row)
        self.txtLL = pack_item(gtk.Entry(), row, expand=True)
        self.txtLL.set_editable(False)
        self.txtLL.set_width_chars(8)
        self.chkUseGPU = pack_check("use GPU", ctrls)
        self.chkUseGPU.set_tooltip_text("If an OpenCL-capable graphics processor is available")
        self.propertied_connect_check('useGPU', self.chkUseGPU)
        self.mnuMethod = pack_item(gtk.combo_box_new_text(), ctrls)
        for i, method_name in enumerate(QUBX_SINGLE_RATES_METHODS):
            self.mnuMethod.append_text(method_name)
            if method_name == self.method:
                self.mnuMethod.set_active(i)
        self.mnuMethod.connect('changed', self.__onChangeMethod)
        self.btnGetLL = pack_button('Get LL', ctrls, on_click=self.__onClickGetLL)
        self.btnAddToTrials = pack_button('Add to Trials', ctrls, on_click=self.__onClickAddToTrials)
        self.btnOpt = pack_button('Optimize Rates', ctrls, on_click=self.__onClickOpt, at_end=True)
        self.optOptions = qubx.optimizeGTK.OptOptionsFace('MILRates', 'qubx_single_rates_optimize', 'QubX.About.MILRates')
        
        self.__data = self.__model = None
        self.__optimizing = False
        self.__onSwitchModel(self.models, self.models.file)
        self.cl_context = None
        
        self.QubX.Modeling.Utils.add_optimizer('MIL', ModelPrep, get_data_prep, self.get_ll_mil, self.optimize_mil, self.OnLL, self.OnIteration, self.OnOptimized)
        self.QubX.Modeling.Utils.add_optimizer('MSL', ModelPrep, get_data_prep, self.get_ll_msl, self.optimize_msl, self.OnLL, self.OnIteration, self.OnOptimized)
    optimizing = property(lambda self: self.__optimizing)
    def propertied_set(self, value, name):
        super(RatesFace, self).propertied_set(value, name)
        self.__gray_ll()
        if name in ('use_group', 'use_group_index'):
            self.plexig.use_group = self.use_group_index if self.use_group else None
        elif name == 'method':
            i = QUBX_SINGLE_RATES_METHODS.index(value)
            self.mnuMethod.set_active(i)
        elif name == 'tdead_in_samples':
            self.__onChangeTdead()
    def onShow(self, showing):
        if showing:
            self.QubX.About.append_face(self.optOptions)
            self.QubX.About.show_face('MILRates')
        else:
            self.QubX.About.remove_face(self.QubX.About.index('MILRates'))
    def qubx_finalize(self):
        self.robot.stop()
        self.data = None
        self.model = None
        #self.QubX.OnQuit -= self.__ref(self.qubx_finalize)
        #self.models.OnSwitch -= self.__ref(self.__onSwitchModel)
    def __onException(self, robot, typ, val, trace):
        traceback.print_exception(typ, val, trace, file=self.outStatus)
    def __onChangeTdead(self):
        td = self.plexig.tdead
        if self.tdead_in_samples:
            self.txtTdead.value = 1e-3*td / self.QubX.Data.file.sampling
        else:
            self.txtTdead.value = td
    def __onEditTdead(self, txt, val):
        if self.tdead_in_samples:
            self.plexig.set_tdead(val * 1e3 * self.QubX.Data.file.sampling, True)
        else:
            self.plexig.set_tdead(val, True)
    def __onSwitchModel(self, models, model):
        if self.__optimizing:
            self.robot.do(gobject.idle_add, self.set_model, model)
        else:
            self.model = model
    def set_model(self, model):
        if model == self.__model: return
        if self.__model:
            for event in (self.__model.states.OnSet,
                          self.__model.states.OnInsert,
                          self.__model.states.OnRemoved,
                          self.__model.rates.OnInsert,
                          self.__model.rates.OnRemoved,
                          self.__model.constraints_kin.OnEdit,
                          self.__model.OnChangeBalanceLoops,
                          self.__model.OnSetChannelCount):
                event -= self.__ref(self.__onSetModel)
            self.__model.rates.OnSet -= self.__ref(self.__onSetRate)
        self.__model = model
        if self.__model:
            for event in (self.__model.states.OnSet,
                          self.__model.states.OnInsert,
                          self.__model.states.OnRemoved,
                          self.__model.rates.OnInsert,
                          self.__model.rates.OnRemoved,
                          self.__model.constraints_kin.OnEdit,
                          self.__model.OnChangeBalanceLoops,
                          self.__model.OnSetChannelCount):
                event += self.__ref(self.__onSetModel)
            self.__model.rates.OnSet += self.__ref(self.__onSetRate)
            self.__onSetModel()
    model = property(lambda self: self.__model, lambda self, x: self.set_model(x))
    def __onInvalidateData(self, product):
        self.model_prep.invalidate()
        self.rates.invalidate()
        self.robot.do(gobject.idle_add, self.__gray_ll)
    def __onInvalidateModel(self, product):
        if not self.__optimizing:
            self.robot.do(gobject.idle_add, self.__gray_ll)
    def __onSetModel(self, *args):
        self.model_prep.invalidate()
        self.rates.invalidate()
    def __onSetRate(self, r, field, val, prev, undoing):
        self.rates.invalidate()
    def __onChangeMethod(self, mnu):
        self.propertied_on_user_set('method', QUBX_SINGLE_RATES_METHODS[mnu.get_active()])
    def __gray_ll(self, gray=True):
        color = gdk.color_parse(gray and "#999999" or "#000000")
        self.txtLL.modify_text(gtk.STATE_NORMAL, color)
    def __begin_task(self, receiver=None):
        self.__receiver = receiver
        self.txtStatus.get_buffer().set_text('')
        Tasks.add_task(self.robot)
        for ctrl in [self.btnGetLL, self.btnAddToTrials, self.btnOpt, self.chkPeq]:
            ctrl.set_sensitive(False)
    def __end_task(self, result):
        try:
            self.LL = result[0]
        except:
            self.LL = result
        Tasks.remove_task(self.robot)
        for ctrl in [self.btnGetLL, self.btnAddToTrials, self.btnOpt, self.chkPeq]:
            ctrl.set_sensitive(True)
        if self.method == 'MIL':
            self.request_show()
        self.__optimizing = False
        if self.__receiver:
            self.__receiver(result)
            del self.__receiver
    def __robot_setup_cl_context(self):
        self.cl_context = msl_accel_context(MSL_ACCEL_OPENCL)
    def __get_model_prep(self):
        if self.__model == None:
            return None
        return ModelPrep(self.__model)
    def __get_rates(self):
        if self.__model == None:
            return None, None
        signal_names = self.QubX.Modeling.Stimulus.active_names
        K0, K1, K2, L, V, P = ModelToRateMatricesP(self.__model, signal_names)
        return K0, K1, K2
    def __set_keys(self, **kw):
        for k, v in kw.iteritems():
            try:
                setattr(self, k, v)
            except:
                print 'MIL/MSL optimizer has no option named %s' % k
    def __onClickGetLL(self, btn):
        qubx.pyenv.env.OnScriptable('QubX.Modeling.MILRates.get_ll()')
        self.get_ll(wait=False)
    def get_ll_mil(self, *args, **kw):
        self.method = 'MIL'
        return self.get_ll(*args, **kw)
    def get_ll_msl(self, *args, **kw):
        self.method = 'MSL'
        return self.get_ll(*args, **kw)
    def get_ll(self, model_prep=None, data_prep=None, wait=True, receiver=lambda LL: None, on_trial=None, **kw):
        if wait:
            return qubx.pyenv.call_async_wait(self.__start_get_ll, model_prep, data_prep, on_trial=on_trial, **kw)[0]
        else:
            self.__start_get_ll(receiver, model_prep, data_prep, on_trial=on_trial, **kw)
    def __start_get_ll(self, receiver=lambda LL: None, model_prep=None, data_prep=None, on_trial=None, **kw):
        self.__begin_task(receiver)
        self.__set_keys(**kw)
        mp = model_prep
        if mp is None:
            mp = self.model_prep.val
            mp.set_rates2(*self.rates.val)
        self.robot.do(self.robot_get_ll, self.use_Peq, mp, data_prep, on_trial)
    def __onGotLL(self, LL, on_trial):
        self.__end_task(LL)
        self.txtLL.set_text('%.6g' % LL)
        self.__gray_ll(False)
        qubx.notebook.Notebook.send(qubx.notebook.NbText('MILRates: Get LL (%s): %.6g' % (self.robot.method, LL)), auto_id='Rates_eval.Messages')
        qubx.notebook.Notebook.send(self.QubX.Models.view.nbPicture, auto_id='Rates_eval.Model.Picture')
        qubx.notebook.Notebook.send(self.QubX.Models.view.nbPictureNO, auto_id='Rates_eval.Model.PictureNO')
        qubx.notebook.Notebook.send(self.QubX.Data.view.hires.nbPicture, auto_id='Rates_eval.Data.Picture')
        qubx.notebook.Notebook.send(self.QubX.Data.view.hires.nbPictureNO, auto_id='Rates_eval.Data.PictureNO')
        qubx.notebook.Notebook.send(self.QubX.Tables.find_view('Rates').nbTable, auto_id='Rates_eval.Model.Rates')
        # publish durhist after any possible durhist.robot activity
        self.QubX.Figures.DurHist.update(force=True)
        self.QubX.Figures.DurHist.robot.do(self.__dhrobot_send_durhists, 'Rates_eval.DurHist.Charts', 'Rates_eval.DurHist.TimeConst')
        self.OnLL(LL)
        if on_trial:
            self.LL = LL
            on_trial(*self.make_trial())
    def __dhrobot_send_durhists(self, auto_id, tc_id):
        gobject.idle_add(self.__send_durhists, auto_id, tc_id)
    def __send_durhists(self, auto_id, tc_id):
        qubx.notebook.Notebook.send(self.QubX.Figures.DurHist.nbCharts, auto_id=auto_id)
        qubx.notebook.Notebook.send(self.QubX.Figures.DurHist.nbTimeConst, auto_id=tc_id)
    def __onClickAddToTrials(self, btn):
        qubx.pyenv.env.OnScriptable('QubX.Modeling.MILRates.add_to_trials(*QubX.Modeling.MILRates.make_trial())')
        self.add_to_trials(*self.make_trial())
    def add_to_trials(self, mtree, row):
        Trials = qubx.global_namespace.QubX.Trials
        ts = Trials.get_trialset(self.method, in_qsf=True)
        ts.append_trial(mtree, row)
        Trials.trialset = ts
    def make_trial(self, fields={}, initial_tree=None, model_prep=None, data_prep=None):
        if model_prep:
            model = model_prep.source or model_prep.model
        else:
            model = self.QubX.Models.file
        if data_prep and data_prep.segs:
            data = data_prep.segs[0].file
        else:
            data = self.QubX.Data.file
        mtree = model.as_tree()
        nParam = fields['nParam'] if ('nParam' in fields) else 0
        row = make_trial_row(model.path, data.path, self.LL, nParam, nData=self.robot.Nevent,
                             nEvent=self.robot.Nevent)
        if initial_tree:
            add_trial_rates(row, initial_tree, 'Initial ')
        add_trial_rates(row, mtree)
        for k,v in fields.iteritems():
            row[k] = v
        return mtree, row
    def __onClickOpt(self, btn):
        qubx.pyenv.env.OnScriptable('QubX.Modeling.MILRates.optimize(on_trial=QubX.Modeling.MILRates.add_to_trials)')
        self.optimize(wait=False, on_trial=self.add_to_trials)
    def optimize_mil(self, *args, **kw):
        self.method = 'MIL'
        return self.optimize(*args, **kw)
    def optimize_msl(self, *args, **kw):
        self.method = 'MSL'
        return self.optimize(*args, **kw)
    def optimize(self, model_prep=None, data_prep=None, wait=True, receiver=lambda LL: None, on_trial=None, **kw):
        if wait:
            return qubx.pyenv.call_async_wait(self.__start_opt, model_prep, data_prep, on_trial=on_trial, **kw)[0]
        else:
            self.__start_opt(receiver, model_prep, data_prep, on_trial=on_trial, **kw)
    def __start_opt(self, receiver=lambda LL: None, model_prep=None, data_prep=None, ext_return=False, on_trial=None, **kw):
        self.iterations = 0
        self.iter_trials = []
        self.grads = []
        self.__on_trial = on_trial
        self.__optimizing = True
        self.__begin_task(receiver)
        self.__set_keys(**kw)
        mp = model_prep
        if mp is None:
            mp = self.model_prep.val
            if mp.source:
                mp.source.undoStack.enable(False)
            mp.set_rates2(*self.rates.val)
        self.__mtree_pre_opt = mp.model.as_tree()
        Trials = qubx.global_namespace.QubX.Trials
        Trials.get_trialset("%s Iterations"%self.method, in_qsf=True).clear()
        self.robot.do(self.robot_opt, self.use_Peq,
                      qubx.optimize.get_opts("qubx_single_rates_optimize",
                                             on_start=lambda rec, opts: gobject.idle_add(self.optOptions.optimizers.onStartOpt, rec, opts),
                                             on_end=lambda rec, opts: gobject.idle_add(self.optOptions.optimizers.onEndOpt, rec, opts)),
                      mp, data_prep, ext_return)
        if self.method == 'MIL':
            self.QubX.Figures.DurHist.request_show()
        self.QubX.Models.view.optimizing = True
    def __onIteration(self, LL, K0, K1, K2, model_prep, grads):
        self.iterations += 1
        self.LL = LL
        if model_prep.source:
            K012ToModel(K0, K1, K2, model_prep.source)
        self.txtLL.set_text('%.6g' % LL)
        self.__gray_ll(False)
        self.outStatus.write('Iter: %i\tLL: %8.5g\tgrads: %s\n' % (self.iterations, LL, ', '.join("%8.5g"%g for g in grads)))
        self.OnIteration(self.iterations, LL, self.grads)
    def __onOptDone(self, LL, workspaces, model_prep, ext_return):
        self.QubX.Models.view.optimizing = False
        self.outStatus.write('Finished optimizing (%s).                                                 LL = %.6g\n' % (self.robot.method, LL))
        try:
            if model_prep.source:
                model_prep.source.undoStack.enable()
                model_prep.source.undoStack.push_undo(bind(model_prep.source.from_tree, self.__mtree_pre_opt),
                                                      bind(model_prep.source.from_tree, model_prep.source.as_tree()))
            self.__set_model_confidence(model_prep, self.robot.K, self.robot.dK)
            if model_prep.source:
                model_prep.source.undoStack.seal_undo('set rates')
            self.txtLL.set_text('%.6g' % LL)
            self.__gray_ll(False)
        finally:
            if ext_return:
                self.__end_task((LL, self.iterations, self.grads, self.hessian))
            else:
                self.__end_task(LL)
        if self.__on_trial:
            mtree, trial_row = self.make_trial({'nParam' : len(self.grads), 'Iterations' : self.iterations},
                                               initial_tree=self.__mtree_pre_opt, model_prep=model_prep, data_prep=self.robot.data_prep)
            for i, g in enumerate(self.grads):
                trial_row['Gradient %i' % i] = g
            self.__on_trial(mtree, trial_row)
            self.__on_trial = None
        qubx.notebook.Notebook.send(qubx.notebook.NbText("""MILRates: Finished optimizing %d events after %d iterations (%s).
Data:      %s
Model:     %s
P0:        %s
LL:        %.6g""" % (self.robot.Nevent, self.robot.iter, self.robot.method,
                      '\n           '.join(os.path.split(ws.view.file.path)[1] for ws in workspaces),
                      model_prep.source and os.path.split(model_prep.source.path)[1] or "(from script)",
                      self.robot.Peq and "equilibrium" or "fixed (States table)",
                      LL)), auto_id='Rates_opt.Messages')
        qubx.notebook.Notebook.send(self.QubX.Models.view.nbPicture, auto_id='Rates_opt.Model.Picture')
        qubx.notebook.Notebook.send(self.QubX.Models.view.nbPictureNO, auto_id='Rates_opt.Model.PictureNO')
        qubx.notebook.Notebook.send(self.QubX.Data.view.hires.nbPicture, auto_id='Rates_opt.Data.Picture')
        qubx.notebook.Notebook.send(self.QubX.Data.view.hires.nbPictureNO, auto_id='Rates_opt.Data.PictureNO')
        qubx.notebook.Notebook.send(self.QubX.Tables.find_view('Rates').nbTable, auto_id='Rates_opt.Model.Rates')
        qubx.notebook.Notebook.send(self.QubX.Tables.find_view('Kinetic Constraints').nbTable, auto_id='Rates_opt.Model.Constraints')
        # publish durhist after any possible durhist.robot activity
        self.QubX.Figures.DurHist.update(force=True)
        self.QubX.Figures.DurHist.robot.do(self.__dhrobot_send_durhists, 'Rates_opt.DurHist.Charts', 'Rates_opt.DurHist.TimeConst')

        ts = qubx.global_namespace.QubX.Trials.get_trialset("%s Iterations"%self.method, in_qsf=True)
        for mtree, row in self.iter_trials:
            ts.append_trial(mtree, row)

        self.OnOptimized(self.iterations, LL, self.grads, self.hessian)
    def __set_model_confidence(self, model_prep, K, dK):
        have_dK = any(dK)
        self.outStatus.write('%10s%7s' % ("Rate From", "To"), scroll=False)
        self.outStatus.write_bold('%14s' % "k0", scroll=False)
        if have_dK:
            self.outStatus.write('%11s' % "+/-", scroll=False)
        self.outStatus.write_bold('%14s' % 'k1', scroll=False)
        if have_dK:
            self.outStatus.write('%11s' % "+/-", scroll=False)
        self.outStatus.write_bold('%14s' % 'k2', scroll=False)
        if have_dK:
            self.outStatus.write('%11s' % "+/-", scroll=False)
        self.outStatus.write('\n')
        model = model_prep.source
        if not model:
            return
        k_ix = K_index(model.states.size)
        rr = model.rates
        for r in xrange(rr.size):
            a, b = [rr.get(r, f) for f in ('From', 'To')]
            i = k_ix.index((a,b))
            self.outStatus.write('%10d%7d' % (a,b), scroll=False)
            self.outStatus.write_bold('%14.5g' % rr.get(r, 'k0'), scroll=False)
            k0_lo = k0_hi = k0_std = 0.0
            if dK[i]:
                k0_lo = float(exp(K[i] - dK[i]))
                k0_hi = float(exp(K[i] + dK[i]))
                k0_std = float(exp(K[i]) * dK[i])
                rr.set(r, 'k0 lo', k0_lo)
                rr.set(r, 'k0 hi', k0_hi)
                rr.set(r, 'k0 std', k0_std)
            if have_dK:
                self.outStatus.write('%11.3g' % k0_std, scroll=False)
            i += len(k_ix)
            self.outStatus.write_bold('%14.5g' % K[i], scroll=False)
            k1_lo = k1_hi = k1_std = 0.0
            if dK[i]:
                k1_lo = float(K[i] - dK[i])
                k1_hi = float(K[i] + dK[i])
                k1_std = float(dK[i])
                rr.set(r, 'k1 lo', k1_lo)
                rr.set(r, 'k1 hi', k1_hi)
                rr.set(r, 'k1 std', k1_std)
            if have_dK:
                self.outStatus.write('%11.3g' % k1_std, scroll=False)
            i += len(k_ix)
            self.outStatus.write_bold('%14.5g' % K[i], scroll=False)
            k2_lo = k2_hi = k2_std = 0.0
            if dK[i]:
                k2_lo = float(K[i] - dK[i])
                k2_hi = float(K[i] + dK[i])
                k2_std = float(dK[i])
                rr.set(r, 'k2 lo', k2_lo)
                rr.set(r, 'k2 hi', k2_hi)
                rr.set(r, 'k2 std', k2_std)
            if have_dK:
                self.outStatus.write('%11.3g' % k2_std, scroll=False)
            self.outStatus.write('\n')
    def robot_on_status(self, msg):
        gobject.idle_add(self.outStatus.write, msg+"\n")
    def robot_setup(self, Peq, model_prep, data_prep):
        self.robot.Peq = Peq
        self.robot.model_prep = model_prep
        self.robot.model = self.robot.model_prep.model
        self.robot.K = K012toK(self.robot.model_prep.K0, self.robot.model_prep.K1, self.robot.model_prep.K2)
        self.robot.clazz = ModelToClazz(self.robot.model)
        self.robot.method = self.method
        if self.robot.Peq:
            self.robot.P0 = None
        else:
            self.robot.P0 = ModelToP0(self.robot.model)
        if self.robot.clazz.shape[0] == 0:
            self.robot.data_prep = Anon(ws=[]) # no model -> do nothing
        elif data_prep:
            self.robot.data_prep = data_prep
        else:
            self.plexig.model_prep = model_prep
            self.robot.data_prep = self.robot.gui_call_recv(self.plexig.request, self.robot)[0]
        self.robot.workspaces = self.robot.data_prep.ws
        self.robot.Nevent = 0
        for ws in self.robot.workspaces:
            try:
                ws.single_model.dispose() # work around callback-related memory leak
            except:
                pass
            ws.single_model = ModelToFast(self.robot.model, ws.sig_names, self.robot_on_status)
            ws.multi_model = qubx.fast.model.MultiModel(ws.single_model)
            ws.multi_model.update()
            if self.robot.Peq:
                ws.P0 = None
            else:
                ws.P0 = buffer_as_array_of_double(ws.multi_model.multi_model[0].P0, ws.multi_model.multi_model[0].Nstate)
            if self.robot.method == 'MIL':
                if len(ws.scls) > 1:
                    gobject.idle_add(self.outStatus.write, "MIL can't analyze data with a changing stimulus.  Please use MSL instead.\n")
                    return False
                elif __builtins__['any'](mc[0] < 0 for mc in ws.mcls):
                    gobject.idle_add(self.outStatus.write, "MIL can't analyze data with blank or excluded regions.  Please extract, or use MSL instead.\n")
                    return False
            self.robot.Nevent += sum(len(durs) for clss, durs in ws.segs)
            if not self.cl_context:
                self.__robot_setup_cl_context()
            if not ws.cl_data:
                ws.cl_data = msl_accel_data(self.cl_context, ws.segs, ws.mcls, ws.scls)
            ws.cl_model = msl_accel_models(self.cl_context, cl_dimension(ws.multi_model.multi_model[0].Nstate), 1, len(ws.scls))
        return True
    def robot_slice_data(self, lbl, segs, mcls, scls, td):
        self.robot.Nevent += sum(len(durs) for clss, durs in segs)
        gobject.idle_add(self.outStatus.write, '%s: %i events\n' % (lbl, self.robot.Nevent))
        return [([seg], mcls, scls) for seg in segs if len(seg[0])]
    def robot_eval(self, K):
        if not self.robot.workspaces:
            gobject.idle_add(self.outStatus.write, "The model is empty.\n")
            return 0.0
        elif not self.robot.data_prep.ws:
            gobject.idle_add(self.outStatus.write, "There are no data files in this group (see the Data table).\n")
            return 0.0
        if self.robot.Nevent == 0:
            gobject.idle_add(self.outStatus.write, "To use this panel, first Idealize your single-molecule data.\nFor ensemble data, try MacRates.\n")
            return 0.0
        LL = 0.0
        K0, K1, K2 = KtoK012(K, self.robot.workspaces[0].single_model.K0)
        argses = []
        iarr, imat, dmat = buffer_as_array_of_int, buffer_as_matrix_of_int, buffer_as_matrix_of_double
        for ws in self.robot.workspaces:
            ws.single_model.K0[:,:] = K0[:,:]
            ws.single_model.K1[:,:] = K1[:,:]
            ws.single_model.K2[:,:] = K2[:,:]
            ws.multi_model.update()
            mm = ws.multi_model.multi_model
            Ns = mm[0].Nstate
            method = self.cl_context.run_mil if (self.robot.method == 'MIL') else self.cl_context.run_msl
            ws.cl_model.reset()
            if ws.cl_model.setup_model(ws.cl_data, iarr(mm[0].clazz, Ns), ws.P0, dmat(mm[0].K0[0], Ns, Ns), dmat(mm[0].K1[0], Ns, Ns), dmat(mm[0].K2[0], Ns, Ns),
                                       imat(mm[0].L[0], Ns, Ns), imat(mm[0].V[0], Ns, Ns), imat(mm[0].P[0], Ns, Ns),
                                       ws.td_ms*1e-3, lambda msg, obj: gobject.idle_add(self.outStatus.write, msg+"\n")):
                return -1e21
            if method(ws.cl_data, ws.cl_model, self.useGPU and MSL_ACCEL_OPENCL or MSL_ACCEL_OPENMP):
                gobject.idle_add(self.disable_gpu)
                if method(ws.cl_data, ws.cl_model, MSL_ACCEL_OPENMP):
                    return -1e20
            LL += ws.cl_model.ll[0]
        return LL
    def disable_gpu(self):
        self.useGPU = False
        self.chkGPU.set_sensitive(False)
        self.chkGPU.set_tooltip_text("OpenCL is not available; needs a better graphics card or new drivers.")
    def robot_get_ll(self, Peq, model_prep, data_prep, on_trial):
        if not self.robot_setup(Peq, model_prep, data_prep):
            gobject.idle_add(self.__onGotLL, 0.0, on_trial)
        else:
            try:
                LL = self.robot_eval(self.robot.K)
                gobject.idle_add(self.__onGotLL, LL, on_trial)
            except:
                gobject.idle_add(self.__end_task, -1.0)
                raise
        del self.robot.workspaces
    def robot_opt(self, Peq, optimize, model_prep, data_prep, ext_return):
        self.robot.optimize = optimize
        self.robot.progress = 0.0
        self.robot.iter = 0
        self.robot.LL = 0.0
        if not self.robot_setup(Peq, model_prep, data_prep):
            if ext_return:
                gobject.idle_add(self.__end_task, (-1.0, 0, [], None))
            else:
                gobject.idle_add(self.__end_task, -1.0)
            return
        try:
            K0, K1, K2 = KtoK012(self.robot.K, self.robot.workspaces[0].single_model.K0)
            A, B = self.robot.model.constraints_kin.get_matrices_p(K0, K1, K2,
                                                                   self.robot.workspaces[0].single_model.L, self.robot.workspaces[0].single_model.V,
                                                                   self.robot.workspaces[0].single_model.P)
            #A, B = reduce_constraints(A, B)
            #self.robot.A, self.robot.B, self.robot.Ainv, pars = linear_constraints(A, B, self.robot.K)
            self.robot.Kr, AinR, BinR, self.robot.Kr_indices = qubx.fast.model.EliminateFixedParams(self.robot.K, A, B)
            A, B = qubx.fast.model.ReduceConstraints(AinR, BinR)
            self.robot.A, self.robot.B, self.robot.Ainv, pars = qubx.fast.model.SetupLinearConstraints(A, B, self.robot.Kr)
            self.robot.Asc, self.robot.Bsc, self.robot.Asci, pars = start_at_ones(pars)
            self.robot.pars = array([x for x in pars.A[:,0]])
            self.robot.npar = len(self.robot.pars)
            if self.robot.npar > 0:
                for ws in self.robot.workspaces:
                    ws.cl_models = msl_accel_models(self.cl_context, cl_dimension(ws.multi_model.multi_model[0].Nstate), 1+2*self.robot.npar, len(ws.scls))
                score, pars, iterations, grads, Hessian = self.robot.optimize(array(self.robot.pars), self.robot_func, self.robot_iter, self.robot_grad_func)
                # make sure final rates are showing (if it overshot and gave up)
                self.robot_iter(self.robot.pars, grads, iterations, - self.robot.LL)
            else:
                score, pars, iterations, grads, Hessian = self.robot_eval(self.robot.K), [], 0, [], numpy.zeros(shape=(0,0))
                self.robot.LL = - score
            dKr = CalcStdOfK(self.robot.A, self.robot.Asc, Hessian)
            self.robot.dK = zeros(shape=self.robot.K.shape, dtype='float64')
            qubx.fast.model.Kr_to_K(dKr, self.robot.dK, self.robot.Kr_indices)
            self.iterations = iterations
            self.grads = grads
            self.hessian = Hessian
        except:
            traceback.print_exc()
            score = -1.0
            if ext_return:
                gobject.idle_add(self.__end_task, -1.0, self.robot.iter, [], None)
            else:
                gobject.idle_add(self.__end_task, -1.0)
            raise
        self.LL = score
        gobject.idle_add(self.__onOptDone, self.robot.LL, self.robot.workspaces, self.robot.model_prep, ext_return)
        del self.robot.workspaces
        del self.robot.model_prep
    def robot_func(self, pars):
        self.robot.Kr = ParsToK(parsToK(array(pars[:self.robot.npar]), self.robot.Asc, self.robot.Asci, self.robot.Bsc), self.robot.A, self.robot.Ainv, self.robot.B)
        qubx.fast.model.Kr_to_K(self.robot.Kr, self.robot.K, self.robot.Kr_indices)
        LL = self.robot_eval(self.robot.K)
        #print -LL,'of',self.robot.K
        return - LL
    def robot_grad_func(self, pars, grads, dx):
        iarr, imat, dmat = buffer_as_array_of_int, buffer_as_matrix_of_int, buffer_as_matrix_of_double
        LL = 0.0
        for ix in xrange(self.robot.npar):
            grads[ix] = 0.0
        valids = []
        xpars = numpy.array(pars[:self.robot.npar], dtype='float64', copy=True)
        for ws in self.robot.workspaces:
            ws.cl_models.reset()
            for ix in xrange(0, self.robot.npar+1):
                deltas = [0] if (ix == self.robot.npar) else [-dx, dx]
                for d in deltas:
                    xpars[:self.robot.npar] = pars[:self.robot.npar]
                    if d:
                        xpars[ix] += d

                    self.robot.Kr = ParsToK(parsToK(xpars, self.robot.Asc, self.robot.Asci, self.robot.Bsc), self.robot.A, self.robot.Ainv, self.robot.B)
                    qubx.fast.model.Kr_to_K(self.robot.Kr, self.robot.K, self.robot.Kr_indices)
                    #print 'q',self.robot.K
                    K0, K1, K2 = KtoK012(self.robot.K, self.robot.workspaces[0].single_model.K0)
                    ws.single_model.K0[:,:] = K0[:,:]
                    ws.single_model.K1[:,:] = K1[:,:]
                    ws.single_model.K2[:,:] = K2[:,:]
                    ws.multi_model.update()
                    mm = ws.multi_model.multi_model
                    Ns = mm[0].Nstate
                    
                    valid = True
                    if ws.cl_models.setup_model(ws.cl_data, iarr(mm[0].clazz, Ns), ws.P0, dmat(mm[0].K0[0], Ns, Ns), dmat(mm[0].K1[0], Ns, Ns), dmat(mm[0].K2[0], Ns, Ns),
                                                imat(mm[0].L[0], Ns, Ns), imat(mm[0].V[0], Ns, Ns), imat(mm[0].P[0], Ns, Ns),
                                                ws.td_ms*1e-3, lambda msg, obj: gobject.idle_add(self.outStatus.write, msg+"\n")):
                        valid = False
                        if not d: # central params must succeed:
                            return 1e20 # usually positive eigenvalues
                    valids.append(valid)
            method = self.cl_context.run_mil if (self.robot.method == 'MIL') else self.cl_context.run_msl
            ws.cl_err = method(ws.cl_data, ws.cl_models, self.useGPU and MSL_ACCEL_OPENCL or MSL_ACCEL_OPENMP)
                #print ws.cl_data.ll
                #print ws.cl_models.ll
                #print LL
            if ws.cl_err:
                gobject.idle_add(self.disable_gpu)
                    # retry no opencl
                ws.cl_err = method(ws.cl_data, ws.cl_models, MSL_ACCEL_OPENMP)
                if ws.cl_err:
                    return 1e20
            LL += ws.cl_models.ll[ws.cl_models.Nmodel-1]
            for ix in xrange(self.robot.npar):
                if valids[2*ix+1] and valids[2*ix]:
                    grads[ix] -= (ws.cl_models.ll[2*ix+1] - ws.cl_models.ll[2*ix]) / (2*dx) # ws.cl_models.Nmodel-1]) / dx # 
                elif valids[2*ix+1]:
                    grads[ix] -= (ws.cl_models.ll[2*ix+1] - ws.cl_models.ll[ws.cl_models.Nmodel-1]) / dx;
                elif valids[2*ix]:
                    grads[ix] -= (ws.cl_models.ll[ws.cl_models.Nmodel-1] - ws.cl_models.ll[2*ix]) / dx;
        return -LL
    def robot_iter(self, pars, grads, iterations, score):
        self.robot.iter = iterations
        self.robot.LL = - score
        self.robot.pars[:] = pars[:self.robot.npar]
        self.robot.progress = (self.robot.iter * 100.0 / MAXITER)
        self.robot.Kr = ParsToK(parsToK(self.robot.pars, self.robot.Asc, self.robot.Asci, self.robot.Bsc), self.robot.A, self.robot.Ainv, self.robot.B)
        qubx.fast.model.Kr_to_K(self.robot.Kr, self.robot.K, self.robot.Kr_indices)
        K0, K1, K2 = KtoK012(self.robot.K, self.robot.workspaces[0].single_model.K0)
        gobject.idle_add(self.__onIteration, - score, K0, K1, K2, self.robot.model_prep, grads)
        mtree, row = self.make_trial({"Iteration": iterations}, model_prep=self.robot.model_prep, data_prep=self.robot.data_prep)
        self.iter_trials.append((mtree, row))


def inter_ll_from_subi_args(args):
    use_lib, clazz, P0, K0, K1, K2, L, V, P, tdead_sec, segs, mcls, scls = args
    constants = scls[0]
    renumber = array([mc[0] for mc in mcls], dtype='int32')
    mil_segs = [(array(renumber[classes], copy=True, dtype='int32'), durations) for classes, durations in segs]
    mil_args = (clazz, P0, K0, K1, K2, L, V, P, constants, tdead_sec, mil_segs)
    return inter_ll_or_0(mil_args)
